#!/usr/bin/perl
#########################################################################
# Copyright (c) 2009 DiBcom S.A., All rights reserved
#########################################################################
#
# File        : $RCSfile: dibvue.pl,v $
# Date        : $Date: 2015/05/21 14:34:52 $
# Version     : $Revision: 1.87 $
# Author      : Stephane ANDRIEUX
# Description : 
# Usage       : 
#
# $Log: dibvue.pl,v $
# Revision 1.87  2015/05/21 14:34:52  sandrieu
# Added a magnify switch in case the resolution/unit is not 1000.
#
# Revision 1.86  2014/06/13 15:22:17  sandrieu
# Added a .db filter for Calibre .db files.
#
# Revision 1.85  2014/06/02 12:03:22  sandrieu
# Added a limit Entry to cap the maximum number of error that can be loaded per Rule. It greatly speed up the loding and display time when Rules have thouthends of errors. Defaults to 1000.
#
# Revision 1.84  2014/03/18 17:04:00  sandrieu
# Added a Not option in the search field to filter out all rule or pads that matches the pattern.
#
# Revision 1.83  2014/02/12 18:03:22  sandrieu
# Added a counter that count the rumber of pads (or rules) that are listed in the left listBox.
#
# Revision 1.82  2014/02/12 17:35:11  sandrieu
# Bug fix search.
#
# Revision 1.81  2014/02/12 15:08:30  sandrieu
# Added a search box to filter rule names or pad names with a regexp.
#
# Revision 1.80  2013/09/25 12:33:53  sandrieu
# Under construction : replaced dbCreateText by add_text but issues when zoomOut.
#
# Revision 1.79  2013/08/14 09:58:40  sandrieu
# Just removed the Panedwindow code under construction.
#
# Revision 1.78  2013/08/14 09:55:18  sandrieu
# Added Panedwindow under construction for backup in CVS. Removed in the next release.
#
# Revision 1.77  2013/05/13 15:30:23  sandrieu
# Moved the Cali buton to an Herc buton as the Calibre mode is used more often.
#
# Revision 1.76  2012/12/17 14:12:02  sandrieu
# Added the possibility to load the .LAYOUT_ERRORS file without specifying a .gds file for Hercules.
#
# Revision 1.75  2011-11-30 14:53:52  sandrieu
# Increased the history to 30.
#
# Revision 1.74  2011-11-09 14:26:55  sandrieu
# Added support of the label option for dfII/Virtuoso as well.
#
# Revision 1.73  2011-08-31 16:59:07  sandrieu
# Added a label option that adds the error name at the bottom left corner of each error shape. Available for EDI only.
#
# Revision 1.72  2011-08-24 10:06:20  sandrieu
# Added a -calibre switch to start the GUI in the Calibre mode. Added a -load switch to load the errors at startup.
#
# Revision 1.71  2011-08-11 08:33:13  sandrieu
# Bug fix but_sel_txt_file.
#
# Revision 1.70  2011-08-10 16:13:11  sandrieu
# Improved dibvue_calibre 2nd for a more robust .db file parsing. Basic regression test done.
#
# Revision 1.69  2011-08-01 17:08:31  sandrieu
# Improvement to support TSMC 40nmLP ratios for the density rules. Still not accurate for rules with 3 ratios.
#
# Revision 1.68  2011-05-11 10:58:15  sandrieu
# Bug fix and minor improvement.
#
# Revision 1.67  2011-03-23 11:22:09  sandrieu
# Cleaning.
#
# Revision 1.66  2011/03/22 16:44:59  sandrieu
# New history browse entry. The user can browse through the last 10 accessed LAYOUT_ERRORS files.
#
# Revision 1.65  2011/01/24 15:40:17  sandrieu
# Fixed a bug. dibvue currently supports only one level of hierarchy in the SREFs. Issue a warning in case of multiple levels of hierarchy an discard the 2nd and below levels.
#
# Revision 1.64  2011/01/24 14:46:38  sandrieu
# Added a check for ANGLE in SREFs, and issue a warning to the user because ANGLEs in SREFs are not supported in dibvue.
#
# Revision 1.63  2011/01/17 15:43:18  sandrieu
# Removed the error count limit in dibvue_gds. Now, all the errors of the GDS are loaded in the dibvue GUI.
#
# Revision 1.62  2011/01/17 10:39:22  sandrieu
# Fixed a bug in the COLROW for AREFs (when starting with a space).
#
# Revision 1.61  2011/01/14 10:01:08  sandrieu
# Fixed a bug when matching the densities.
#
# Revision 1.60  2011/01/14 08:52:44  sandrieu
# Improved dibvue_gds to support hierarchical errors in subcells. Now supports Hercules SREF and AREF (without rotation nor miroring). Error count may be higher in the GDS than in the txt file due to arrays AREFs.
#
# Revision 1.59  2011/01/13 17:41:06  sandrieu
# Under construction: Improved dibvue_gds .
#
# Revision 1.58  2011/01/06 17:04:18  sandrieu
# Improved dibvue_calibre for a more robust .db file parsing. Basic regression test done.
#
# Revision 1.57  2010/11/09 15:28:00  sandrieu
# Added support for multiple rule highlighting at once through the --Hi rule-- button. The scroled selection mode has been changed to extended for the rules listBox.
#
# Revision 1.56  2010/04/19 17:29:34  sandrieu
# Implemented the --pan-- feature for dfII/Virtuoso as well.
#
# Revision 1.55  2010/04/19 16:39:45  sandrieu
# Added a --pan-- button to pan to an error instead of zooming to it. Implemented for EDI only so far.
#
# Revision 1.54  2010/04/15 16:31:02  sandrieu
# Bug fix to print the summary for the last error as it is not (always?) displayed.
#
# Revision 1.53  2010/04/15 15:28:21  sandrieu
# Added a warning messsage to inform about that the Ratio filter feature is disabled in case of error limit 1000 found.
#
# Revision 1.52  2010/02/11 16:00:18  sandrieu
# Fixed a window resize issue on some machines. Now uses the /home/users/sandrieu env. var. for the user conf file.
#
# Revision 1.51  2010/02/11 10:56:55  sandrieu
# Added the use of the DIB_HOME env. var. Optionnal if GDS2.pm package file is in the local directory.
#
# Revision 1.50  2010/02/09 17:16:55  sandrieu
# Modification to enable multiple dibVue opened simultaneously on different layout views in dfII.
#
# Revision 1.49  2010/02/09 15:22:40  sandrieu
# Fixed bug in the rule description display.
#
# Revision 1.48  2010/02/05 16:32:58  sandrieu
# Added the /home/users/asictool/EDA use lib path for GDS2.pm
#
# Revision 1.47  2010/02/04 14:00:07  sandrieu
# Bug fix. Forget initialisation sequence for dfII.
#
# Revision 1.46  2010/02/04 11:46:47  sandrieu
# The Gds file entry is now disabled when the cali button is ON. It is more understandable for the user.
#
# Revision 1.45  2010/02/03 18:32:45  sandrieu
# Virtuoso integration. dibvue is now compatible with Virtuoso! To be used along with dibvue.il . load(./SCRIPTS/dibvue.il) .
#
# Revision 1.44  2010/02/02 10:02:59  sandrieu
# Added support for FIND_SHORTEST_PATH_BETWEEN_TEXT_SHORTS errors.
#
# Revision 1.43  2010/01/22 11:34:39  sandrieu
# Improved the listboxes browsing. Ability to browse the Rules listBox with the Up and Down keys.
#
# Revision 1.42  2010/01/21 15:16:02  sandrieu
# Added the ability to display sub hier cells errors missing in the top cell GDS when their bbox is defined for the top cell in the TXT error file.
#
# Revision 1.41  2010/01/21 11:30:17  sandrieu
# Added a limit to max 1000 errors displayed in dibVue per rule type. (hard coded).
#
# Revision 1.40  2009/11/12 17:25:56  sandrieu
# Added a features that remembers the last tx and gds files loaded if no -txt and/or -gds arguments are given on the command line.
#
# Revision 1.39  2009/11/10 14:33:58  sandrieu
# Added a flag -*- to emphasis the discrepancies in the hercules shortSummary.
#
# Revision 1.38  2009/11/10 14:20:24  sandrieu
# The shortSummary for Hercules now has to errors numbers per rule: total count (including sub hier cells) and top cell count. Warning there can be discrepancies.
#
# Revision 1.37  2009/11/10 14:00:18  sandrieu
# Added a shortSummary information for hercules. Fixed a small bug.
#
# Revision 1.36  2009/11/09 15:36:38  sandrieu
# Improved the -clear All- procedure so that it clears also the custom lines stored in the fedb in a previous session.
#
# Revision 1.35  2009/11/06 17:22:36  sandrieu
# The -Hi rule- button is now independant of the status of the -auto Zoom- button.
#
# Revision 1.34  2009/11/05 16:39:35  sandrieu
# Fixed a bug for the 4 lines at the end of a Calibre db file.
#
# Revision 1.33  2009/11/05 14:57:03  sandrieu
# Added support for 0.18um rule name that includes the keyword Rule at the beginning of the rule name.
#
# Revision 1.32  2009/11/04 14:24:27  sandrieu
# Fixed a bug when re-loading a db file for Calibre.
#
# Revision 1.31  2009/11/04 11:03:34  sandrieu
# Added the possibility to load the compare_gds results from Calibre. The layers having 0 differences are filtered but listed while loading.
#
# Revision 1.30  2009/11/03 16:32:39  sandrieu
# Changed the var SOCKET_FILE to SOCKET_PORT. Changed the socket warning message.
#
# Revision 1.29  2009/11/03 15:31:10  sandrieu
# Beta. Added a Calibre DRC and Shorts .db parser. dibvue can now be used as Mentor RVE for DRC errors and Shorts errors.
#
# Revision 1.28  2009/10/12 14:40:02  sandrieu
# Beta. The rules are now sorted in the alphabetical order. Added OFFGRID and SKEW rules. Removed useless code.
#
# Revision 1.27  2009/10/12 12:45:14  sandrieu
# Beta. Removed useless code from the previous socket.
#
# Revision 1.26  2009/10/12 10:00:56  sandrieu
# Beta. New socket system. Replaced the custom socket by the standard tcl/perl/unix socket.
#
# Revision 1.25  2009/10/09 10:16:19  sandrieu
# Beta. Minor change.
#
# Revision 1.24  2009/10/09 08:51:50  sandrieu
# Beta. Fixed socket issue when parsing the text file (while in density matching).
#
# Revision 1.23  2009/10/09 08:22:56  sandrieu
# Beta. Bug fix. The displayed ratio error for a violation was potentially bad.
#
# Revision 1.22  2009/10/08 16:12:04  sandrieu
# Beta. Added an -ascii option to output the GDS in a text format.
#
# Revision 1.21  2009/10/07 14:31:23  sandrieu
# Beta. The Hi rule button now takes in account the status of the ratio filter button and it displays only the listed filtered errors.
#
# Revision 1.20  2009/10/06 17:04:15  sandrieu
# Beta. Temp release for but_clearAll improvement. TBD.
#
# Revision 1.19  2009/10/05 08:36:14  sandrieu
# Beta. Fixed a bug when re-loading new errors in the same session.
#
# Revision 1.18  2009/10/01 16:34:21  sandrieu
# Beta. Fixed a bug: when the filtering feature was used the errors displayed were the wrong ones.
#
# Revision 1.17  2009/10/01 15:34:09  sandrieu
# Beta. Added a ratio filtering feature (with a ratio filter option button) to filter the density errors based on a user specified ratio threshold.
#
# Revision 1.16  2009/10/01 10:13:43  sandrieu
# Beta. Added Current rule background selection color, added rule name in rule description, added min and max ratio for density rules at the end of the description field.
#
# Revision 1.15  2009/09/30 13:18:57  sandrieu
# Alpha. Improved Up and Down arrow behavior. Bug fix for the clear all button.
#
# Revision 1.14  2009/09/29 17:32:44  sandrieu
# Alpha. Added the possibility to use the Up and Down arrows to browse the errors. Warning, bug: the error list needs to be clicked first. To be fixed.
#
# Revision 1.13  2009/09/29 16:08:35  sandrieu
# Alpha. Improved the socket interface while reading the GDS.
#
# Revision 1.12  2009/09/23 13:19:36  sandrieu
# Added comments; Improved the socket interface.
#
# Revision 1.11  2009/09/17 16:24:26  sandrieu
# Alpha. Added the version number in the title.
#
# Revision 1.10  2009/09/17 16:19:20  sandrieu
# Alpha. Improved the GUI resize behavior.
#
# Revision 1.9  2009/09/17 15:49:32  sandrieu
# Alpha. Added the error number and the error ratio information in the errror list listBox.
#
# Revision 1.8  2009/09/17 12:34:13  sandrieu
# Alpha. Added a clear all button.
#
# Revision 1.7  2009/09/17 09:33:00  sandrieu
# Alpha. Added a socket file name and a flag system to allow SOC Encounter to close the socket.
#
# Revision 1.6  2009/09/16 18:55:16  sandrieu
# Alpha. First working release. Requires dibvue.tcl in SOC and a temp file. Button Hi_All not yet functionnal.
#
# Revision 1.5  2009/09/14 13:39:01  sandrieu
# Under construction. Added Error description text and started custom line generation.
#
# Revision 1.4  2009/09/11 16:20:26  sandrieu
# Under construction. Added listBox for error types and error list.
#
# Revision 1.3  2009/09/10 17:42:15  sandrieu
# Under construction. Added the first graphical elements.
#
# Revision 1.2  2009/09/09 15:48:31  sandrieu
# Under construction. Added txt parsing and unified database for gds and txt.
#
# Revision 1.1  2009/09/08 18:42:19  sandrieu
# Under construction.
#
#########################################################################
my $revision = '$Revision: 1.87 $';
my $version;
if ( $revision =~ m/: (\S+) \$/ ) { $version = $1};

if ($ENV{"DIB_HOME"} eq "") {print STDERR "-W- DIB_HOME Environment variable is not set\n";}

use Socket;
use lib ".";
use lib $ENV{"DIB_HOME"};

my %dbase;
my %ENVI;
my @rule_sel;
my @rule_list;
my @rule_list_search;
my %rule_list_index;
my $search_pattern;
my @rule_layer_list;
my @err_list;
my $err_name;
my $err_desc;
my $layer_pair;
my $BUT_Not;
my $BUT_label;
my $BUT_autoZoom;
my $BUT_pan;
my $BUT_hercules;
my $BUT_ratioFilter;
my $lb_rule_list;
my $cur_rule_layer;
my $history_file;
my $dfII;
my $hercules;
my $load;
my $magnify=1;

###############################################################################
#
# Argument checking and help.
#
###############################################################################
use Getopt::Long;

my $nb_arg = scalar(@ARGV);
GetOptions( 
            "socket=s"    => \$ENVI{SOCKET_PORT},
            "t|txt=s"     => \$ENVI{TXT_FILE},
            "g|gds=s"     => \$ENVI{GDS_FILE},
            "a|ascii=s"   => \$ascii,
            "d|dfII"      => \$dfII,
            "h|hercules"  => \$hercules,
            "l|load"      => \$load,
            "m|magnify=s" => \$magnify,
            "h|help"      => \$help,
          );

if ( $help ) {
  die "Usage: $0 -top <top_cell_name> -txt <*.LAYOUT_ERRORS> -gds <EV_OUT.gds>

  -txt <*.LAYOUT_ERRORS|*.db> : Text file generated by Hercules or Calibre. It contains
                                the DRC errors in a text format.
  -gds <EV_OUT.gds>           : GDS file generated by Hercules containing the detail
                                shapes of the DRC errors.
  -ascii <file.txt>           : Writes out the GDS in a text format in <file.txt>.
  -dfII                       : Activates the dfII/Virtuoso/Skill mode.
  -hercules                   : Activates the Hercules mode.
  -load                       : If input file(s) are given in argument (-txt [-gds]) then
                                this option execute the \"Load\" action. 

  -help                       : This message.
\n\n";
}


########################################################
# 
# Checks the -txt and -gds options when -load is used.
# 
########################################################
if ( $load == 1 ) {
  if  ( $ENVI{TXT_FILE} eq "" ) { die ("-E- Error : \"-load\" requires the \"-txt\" switch !\n\n") };
  if  (!(-f $ENVI{TXT_FILE})  ) { die ("-E- Error : The TXT file $ENVI{TXT_FILE} doesn't exist !\n\n") };
  if ( $hercules == 1 ) {
    if  ( $ENVI{GDS_FILE} eq "" ) { die ("-E- Error : \"-load\" requires the \"-gds\" switch !\n\n") };
    if  (!(-f $ENVI{GDS_FILE})  ) { die ("-E- Error : The GDS file $ENVI{GDS_FILE} doesn't exist !\n\n") };
  }
}


########################################################
# 
# Stores the last error file name in the ~/.dibvue file.
# 
########################################################
$history_file = $ENV{HOME} . "/.dibvue";

if ( ($ENVI{TXT_FILE} eq "") && (-f $history_file) ) {
  my $txt_tmp = `grep -m 1 "^txt " $history_file`;
  if ( $txt_tmp =~ /^txt\s+(\S+)$/ ) { $ENVI{TXT_FILE} = $1 }
  &browse_txt_file_sub;
}


########################################################
# 
# Initializing the dfII mode.
# 
########################################################
my $skilltmp;
my $uid;
my $pid;
if ( $dfII ) {
  #Activating the autoflush.
  $|=1;
  $uid = $<;
  $pid = $$;
  $skilltmp = "./.dibvue_tmp_${uid}_${pid}.il";
  print STDERR "-I- Starting dibvue ...\n";
  unlink(<./.dibvue_tmp_*>);

  print "cv${uid}_${pid}=deGetEditCellView() ";
  print "cw${uid}_${pid}=hiGetCurrentWindow() ";
  print "hlSetId = geCreateHilightSet(cv${uid}_${pid} \'(\"y1\" \"drawing\")) ";
  print "hlSetId->enable = t ";

}

########################################################
# 
# Converting the GDS2 file in the ascii format.
# 
########################################################
if ( $ascii ne "" ) {
  
  if  ( $ENVI{GDS_FILE} eq "" ) { die ("-E- Error : please use the -gds switch !\n\n") };
  if  (!(-f $ENVI{GDS_FILE})  ) { die ("-E- Error : The GDS file $ENVI{GDS_FILE} doesn't exist !\n\n") };
  print STDERR "-I- Writing the GDS to the text file : $ascii\n";

  use GDS2;
  my $gds2File = new GDS2(-fileName=>$ENVI{GDS_FILE});

  open(FILE_OUT, ">$ascii") || die ("-E- Error : Can't open txt file $ascii !\n\n");
  while ($gds2File -> readGds2Record) {
    print FILE_OUT $gds2File -> returnRecordAsString . "\n";
  }
  close(FILE_OUT) || die "-E- Can't close $ascii !";

  exit;
}

if ( ($ENVI{SOCKET_PORT} eq "") && (! $dfII) ) {
  print STDERR "-W- No socket port specified!!! dibvue won't work inside SOC/EDI.\n";
}

########################################################
# 
# procedure to call the standard unix socket connection
# 
########################################################
sub get_connection {

  # Called with a host and port. Sets up a TCP connection.

  my ($host,$port) = @_;
  $port ||= 9900;
  my $iaddr = inet_aton($host);
  my $paddr = sockaddr_in($port,$iaddr);
  my $proto = getprotobyname("tcp");
  socket (SOCK, PF_INET, SOCK_STREAM, $proto)
          or die ("no free sockets\n");
  connect (SOCK, $paddr)
          or die ("$host not responding\n");
}

###############################################################################
#
# Building the GUI.
#
###############################################################################
use strict;
use Tk;
use Tk::BrowseEntry;
use Tk::Wm;
use Tk::NoteBook;
use Tk::ProgressBar;
require Tk::LabFrame;
require Tk::Dialog;
require Tk::DialogBox;

my $tk_version = $Tk::VERSION;

#xcolors gcolorsel donnent le nom de chaque couleur
my $bg = "grey85";
my $fg = "black";
my $selColor = "Maroon";
my $disable_fg = "grey70";
####### Use xfontsel to select fonts
#my $font_name="-misc-fixed-medium-r-normal-*-13-120-75-75-*-*-iso8859-1";
my $font_name="AdobeCourier-9";

my $main = new MainWindow(-background => "$bg", -borderwidth => '10' );
$main->geometry("+20+40");
$main->resizable(1,1);
#$main->geometry("465x860+20+40");
#$main->maxsize(465,860);
$main->minsize(500,310);
$main->optionAdd("*background","$bg");
$main->optionAdd("*font","$font_name");
$main->optionAdd("*highlightThickness","0");
$main->optionAdd("*activeBackground","grey95");
$main->optionAdd("*activeForeground","$fg");

########################################################
# 
# GUI actions
# 
########################################################

$ENVI{EDITOR}="nedit";

my $top_frame = $main->Frame(
    -borderwidth => '2', 
    )->pack(-side => 'top');


########################################################
# 
# Title and Txt and Gds input files.
# 
########################################################
my $label1_main = $top_frame->Label(-text => "Hercules & Calibre DRC errors GUI for SOC/EDI & Virtuoso (v$version)", 
    -font => "$font_name", -borderwidth => '0',
    )->pack(-side => 'top');

my $txt_frame = $main->Frame(
    )->pack(-side => 'top', -anchor => 'w',-expand => 'no', -fill => 'x');

my $browse_txt_file = $txt_frame->BrowseEntry(
   -variable => \$ENVI{TXT_FILE}, -label=>"Txt file:", -listwidth=> "800",
   -browsecmd => [\&browse_txt_file_sub]
   )->pack(-side => 'left', -anchor => 'center', -expand => '1', -fill => 'x');
$browse_txt_file->configure(-state=>"normal", -listcmd => [\&fill_history, $browse_txt_file]);
$browse_txt_file->bind('<KeyRelease>' => [\&browse_txt_file_sub]);
$browse_txt_file->bind('<Leave>'      => [\&browse_txt_file_sub]);

my $but_edit_txt_file = $txt_frame->Button(
        -text => "<>", -width => "0", -pady => "0p", -padx => "2p",
        -font => "$font_name",
        )->pack(-side => 'right', -anchor => 'center', -expand => '0', -fill => 'none');
$but_edit_txt_file->configure(-command => [\&but_editFile , $but_edit_txt_file, \$ENVI{TXT_FILE}]);

my $but_sel_txt_file = $txt_frame->Button(
    -text => "...", -width => "0", -pady => "0p", -padx => "2p",
    -font => "$font_name",
    -command => sub {  if ($ENVI{TXT_FILE} eq "") {$ENVI{TXT_FILE} = `pwd`; chomp($ENVI{TXT_FILE})}
                       my $tmp=scriptDialog($main, $browse_txt_file, $ENVI{TXT_FILE});
                       if (-f $tmp) { $ENVI{TXT_FILE}=$tmp; &browse_txt_file_sub };
                    }
    )->pack(-side => 'right');


sub browse_txt_file_sub {
    my $tmp = $ENVI{TXT_FILE};
    $tmp =~ s/\.LAYOUT_ERRORS$/.gds/;
    if ( (-f $tmp) && ( $tmp =~ /\.gds/ ) ) { $ENVI{GDS_FILE} = $tmp; }
} # sub browse_txt_file_sub

sub fill_history {
    my ($widget) = @_ ;
    my @data_list;
    my $data;
    
    
    if ( -f $history_file ) {
      open(FILE_HISTORY, $history_file) || die ("Error : Can't open the history file $history_file !\n\n");
      while(<FILE_HISTORY>) {
        if ( /^txt\s+(\S+)$/ ) { my $txt_tmp = $1; push(@data_list, $txt_tmp ); }
      }
      close(FILE_HISTORY) || die "Can't close $history_file !";
    }

    $widget->delete("0","end");
    foreach $data (@data_list) {
        chomp($data);
        $widget->insert("end", "$data");
    }
} # sub fill_history




my $gds_frame = $main->Frame(
    )->pack(-side => 'top', -anchor => 'w',-expand => 'no', -fill => 'x');

my $label_gds = $gds_frame->Label(-text => "Gds file:", 
    -font => "$font_name", -borderwidth => '0',
    )->pack(-side => 'left');

my $gds_file = $gds_frame->Entry(
    -textvariable => \$ENVI{GDS_FILE} , -width => 50, 
    )->pack(-side => 'left',-expand => 'yes', -fill => 'x');

my $checkBut_txt_file = $gds_frame->Checkbutton(
    -text => "Herc" , -variable => \$BUT_hercules,
    -onvalue => 1, -offvalue => 0, -selectcolor => "$selColor",
    )->pack(-side => 'right');

my $but_sel_gds_file = $gds_frame->Button(
    -text => "...", -width => "0", -pady => "0p", -padx => "2p",
    -font => "$font_name",
    -command => sub { if ($ENVI{GDS_FILE} eq "") {
                        if ($ENVI{TXT_FILE} ne "") { $ENVI{GDS_FILE} = $ENVI{TXT_FILE}; $ENVI{GDS_FILE} =~ s/(.*\/).*/\1/;}
                        else {$ENVI{GDS_FILE} = `pwd`; chomp($ENVI{GDS_FILE}) }
                      }
                      scriptDialog($main, $gds_file, $ENVI{GDS_FILE});
                    }
    )->pack(-side => 'right');

########################################################
# 
# Command frame : Load, clear all, Highlight  rule,
# auto zoom, ratio filter buttons.
# 
########################################################
my $cmd_frame = $main->Frame(
    )->pack(-side => 'top', -anchor => 'w');

my $but_load = $cmd_frame->Button(-text => 'Load',
    -command => \&but_load_action , 
    )->pack(-side => 'left', -anchor => 'w');

my $but_clearAll = $cmd_frame->Button(-text => 'Clear All',
    -command => \&but_clearAll , 
    )->pack(-side => 'left', -anchor => 'w');

my $but_hiRule = $cmd_frame->Button(-text => 'Hi rule',
    -command => \&but_hi_rule , 
    )->pack(-side => 'left', -anchor => 'w');

my $checkBut_label = $cmd_frame->Checkbutton(
    -text => "label" , -variable => \$BUT_label,
    -onvalue => 1, -offvalue => 0, -selectcolor => "$selColor",
    )->pack(-side => 'left', -anchor => 'w');

my $checkBut_autoZoom = $cmd_frame->Checkbutton(
    -text => "auto Zoom" , -variable => \$BUT_autoZoom,
    -onvalue => 1, -offvalue => 0, -selectcolor => "$selColor",
    )->pack(-side => 'left', -anchor => 'w');
$checkBut_autoZoom->configure(-command => sub{ if ( $BUT_autoZoom == 1 ) {$BUT_pan=0;} } );

my $checkBut_pan = $cmd_frame->Checkbutton(
    -text => "pan" , -variable => \$BUT_pan,
    -onvalue => 1, -offvalue => 0, -selectcolor => "$selColor",
    )->pack(-side => 'left', -anchor => 'w');
$checkBut_pan->configure(-command => sub{ if ( $BUT_pan == 1 ) {$BUT_autoZoom=0;} } );

my $checkBut_ratioFilter = $cmd_frame->Checkbutton(
    -text => "ratio Filter" , -variable => \$BUT_ratioFilter,
    -onvalue => 1, -offvalue => 0, -selectcolor => "$selColor",
    )->pack(-side => 'left', -anchor => 'w');

my $ent_ratioFilter = $cmd_frame->Entry(
    -textvariable => \$ENVI{RATIO_FILTER} , -width => 4, 
    )->pack(-side => 'left',-expand => 'no', -fill => 'none');

#$ent_ratioFilter->bind('<KeyRelease>' => sub{ &set_env("LOG_FILE", \$ENVI{"LOG_FILE"}, \${ENV_FILE}); });
#$ent_ratioFilter->bind('<Leave>'      => sub{ &set_env("LOG_FILE", \$ENVI{"LOG_FILE"}, \${ENV_FILE}); });

$checkBut_ratioFilter->configure(-command => sub{ 
                                         if ( $BUT_ratioFilter == 1 ) {
                                           $ent_ratioFilter->configure(-state=>"normal", -foreground => "$fg");
                                           $ent_ratioFilter->configure(-state=>"normal")
                                         } else {
                                           $ent_ratioFilter->configure(-state=>"disabled", -foreground => "$disable_fg");
                                           $ent_ratioFilter->configure(-state=>"disabled")
                                         }
                                        }
                       );

$checkBut_ratioFilter->invoke;
$checkBut_ratioFilter->invoke;

sub wait_dfII {
  my $timeout=0;
  while( (-f $skilltmp) && ($timeout <= 100) ) {
    $timeout++;
    select(undef, undef, undef, 0.001);
    if ($timeout == 100) {
      print STDERR "-W- Internal error! Timout for the skill tmp file $skilltmp .\n";
      system "rm $skilltmp";
    }
  }
}

sub but_load_action {
  
  #Reset the internal perl error database.
  %dbase = ();
  @rule_list = ();
  @err_list = ();
  my $j;

  if (-f $ENVI{TXT_FILE}) {system "echo 'txt $ENVI{TXT_FILE}' >> ${history_file}.tmp" };
  open(FILE_HISTORY, $history_file) || die ("Error : Can't open the history file $history_file !\n\n");
  my $l_count=1;
  while(my $line=<FILE_HISTORY>) {
    chomp($line);
    if ( !($line=~/$ENVI{TXT_FILE}/) && ($l_count++ < 31) ) { system "echo '$line' >> ${history_file}.tmp" }
  }
  close(FILE_HISTORY) || die "Can't close $history_file !";
  system "mv ${history_file}.tmp $history_file";

  #if (-f $ENVI{GDS_FILE}) {system "echo 'gds $ENVI{GDS_FILE}' >> $history_file"};

  if ($BUT_hercules == 1 ) {
    if (-f $ENVI{GDS_FILE}) {&dibvue_gds} else { print STDERR "-E- GDS file doesn't exist : $ENVI{GDS_FILE}\n\n"; };
    if (-f $ENVI{TXT_FILE}) {&dibvue_txt} else { print STDERR "-E- TXT file doesn't exist : $ENVI{TXT_FILE}\n\n"; };
  } else {
    if (-f $ENVI{TXT_FILE}) {&dibvue_calibre} else { print STDERR "-E- TXT file doesn't exist : $ENVI{TXT_FILE}\n\n"; };
  }
  
  ########################################################
  # 
  # Sorting the rules layer list based on the alphabetical order
  # of the rule names.
  # 
  ########################################################
  my @tmp;
  my $rule_layer;
  my @rule_nb;
  @rule_nb = values(%{$dbase{$ENVI{TOP_CELL}}{rule_nb}});
  @rule_layer_list = keys(%{$dbase{$ENVI{TOP_CELL}}{rule_nb}});
  @rule_nb = sort (@rule_nb);
  for ($j = 0; $j <= $#rule_nb; $j++) {
    foreach $rule_layer (@rule_layer_list) {
      if ( $dbase{$ENVI{TOP_CELL}}{rule_nb}{$rule_layer} eq $rule_nb[$j] ) { push(@tmp, $rule_layer ); }
    }
  }
  @rule_layer_list=@tmp;
  
  ########################################################
  # 
  # Building the displayed rule list with the number of
  # error for each rule.
  # 
  ########################################################
  for ($j = 0; $j <= $#rule_layer_list; $j++) {
    my $count = $#{$dbase{$ENVI{TOP_CELL}}{geom}{@rule_layer_list[$j]}} + 1;
    push(@rule_list, "$dbase{$ENVI{TOP_CELL}}{rule_nb}{@rule_layer_list[$j]} ($count)" );
  }
  
  @rule_list_search = @rule_list;
  @rule_list_index{@rule_list} = (0..$#rule_list);
  $ENVI{SEARCH_COUNT} = scalar(@rule_list);

}

sub but_clearAll {

  if (! $dfII ) {
    &get_connection("localhost", $ENVI{SOCKET_PORT});
    print SOCK "if ![info exists dibvue_line_list] {set dibvue_line_list {}}\n";
    print SOCK "foreach dibvue_line_ptr \$dibvue_line_list {dbDeleteCustomObj \$dibvue_line_ptr}; redraw\n";
    print SOCK "unset dibvue_line_list\n";
  
    # Clears all custom lines in the db. Removes the need for dibvue_line_list var.
    # Works as well on a re-loaded database (erease sotres custom lines in the fedb).
    print SOCK "deselectAll\n";
    print SOCK "dbForEachFPlanGeomList [dbHeadFPlan] g { dbForEachGeomListGeom \$g ptr {dbSelectObj \$ptr} }\n";
    print SOCK "deleteSelectedFromFPlan\n";
    print SOCK "dbDeleteText -all\n";
  
    close(SOCK) || die "Can't close Socket !";
  } else {
    print "geDeleteAllHilightSet(cv${uid}_${pid}) ";
    print "hlSetId = geCreateHilightSet(cv${uid}_${pid} \'(\"y1\" \"drawing\")) ";
    print "hlSetId->enable = t ";
    print "hiRedraw(cw${uid}_${pid}) ";
  }

}

sub but_hi_rule {
  
  my $j;
  my $k;
  my $error_array;
  my $coord_string;
  my $ratio;

  &wait_dfII;

  if (! $dfII ) {&get_connection("localhost", $ENVI{SOCKET_PORT});}
  else          {open (SKILLTMP, ">$skilltmp") || die ("*** Error: Can't open $skilltmp !\n\n");}


  #- Foreach rule selected, several rules can be selected at once.
  for ($k = 0; $k <= $#rule_sel; $k++) {
    $cur_rule_layer = @rule_layer_list[@rule_sel[$k]];
    
    #- Foreach error of the current rule, display it.
    for ($j = 0; $j <= $#{$dbase{$ENVI{TOP_CELL}}{geom}{$cur_rule_layer}}; $j++) {
      $ratio=$dbase{$ENVI{TOP_CELL}}{ratio}{$cur_rule_layer}[$j] * 100;
      if      ( !($BUT_ratioFilter) || (($ENVI{RATIO_FILTER} >= 50) && ($ratio >= $ENVI{RATIO_FILTER})) ) {
        &disp_errors($j,1);
      } elsif ( !($BUT_ratioFilter) || (($ENVI{RATIO_FILTER}  < 50) && ($ratio <= $ENVI{RATIO_FILTER})) ) {
        &disp_errors($j,1);
      }
    }
  }

  if (! $dfII ) {print SOCK "redraw\n";}

  if (! $dfII ) {close(SOCK) || die "Can't close Socket !";}
  else          {close(SKILLTMP) || die "Can't close $skilltmp !";}

  if ( $dfII ) {print "hiRedraw(cw${uid}_${pid}) load(\"$skilltmp\") system(\"rm $skilltmp\") ";}

}


########################################################
# 
# Load limit to to speed up the loading and display time.
# Setting a default of 1000 errors max. per rule.
# 
########################################################
$ENVI{LIMIT} = 1000;

my $limit_search_frame = $main->Frame(
    )->pack(-side => 'top', -anchor => 'w', -expand => 'no', -fill => 'x');

my $label_limitBox = $limit_search_frame->Label(
    -text => "Limit:", -width => 6, -anchor => 'w'
    )->pack(-side => 'left',-expand => 0, -fill => 'none');

my $entry_limitBox = $limit_search_frame->Entry(
    -textvariable => \$ENVI{LIMIT} , -width => 7
    )->pack(-side => 'left',-expand => 0, -fill => 'none');

########################################################
# 
# searchBox to filter the rules containing errors.
# 
########################################################
my $label_searchBox = $limit_search_frame->Label(
    -text => "Search:", -width => 8, -anchor => 'e'
    )->pack(-side => 'left',-expand => 0, -fill => 'none');
my $checkBut_searchInv = $limit_search_frame->Checkbutton(
    -text => "Not" , -variable => \$BUT_Not,
    -onvalue => 1, -offvalue => 0, -selectcolor => "$selColor",
    )->pack(-side => 'left', -anchor => 'w');
$checkBut_searchInv->configure(-command => sub{ &but_search_action(\$ENVI{SEARCH}); } );
my $entry_searchBox = $limit_search_frame->Entry(
    -textvariable => \$ENVI{SEARCH} , -width => 20
    )->pack(-side => 'left',-expand => 1, -fill => 'x');
$entry_searchBox->bind('<KeyRelease>' => sub{ &but_search_action(\$ENVI{SEARCH}); } );
#$labent_searchBox->eventAdd('<<Paste>>' => '<Control-v>'); 
#$labent_searchBox->eventAdd('<<Paste>>' => '<Button-2>'); 

my $label_search_count = $limit_search_frame->Entry(
    -textvariable => \$ENVI{SEARCH_COUNT}, -borderwidth => '0', -width => 5,
    )->pack(-side => 'right', -expand => 'no', -fill => 'none');


sub but_search_action {
  my ($var) = @_ ;
  $search_pattern = $$var;
  
  my $rule_tmp;
  my @rule_list_tmp = ();
  if ($search_pattern eq "" ) {
    @rule_list_tmp = @rule_list
  } else {
    foreach $rule_tmp (@rule_list) {
      if ( $BUT_Not == 0 ) {
        if ($rule_tmp =~ /$search_pattern/) { push(@rule_list_tmp, $rule_tmp); }
      } else {
        if ($rule_tmp !~ /$search_pattern/) { push(@rule_list_tmp, $rule_tmp); }
      }
    }
  }
  @rule_list_search = @rule_list_tmp;
  $ENVI{SEARCH_COUNT} = scalar(@rule_list_search);
}


$checkBut_txt_file->configure(-command => sub{ 
                                         if ( $BUT_hercules == 1 ) {
                                           $label_gds->configure(-state=>"normal", -foreground => "$fg");
                                           $gds_file->configure(-state=>"normal", -foreground => "$fg");
                                           $but_sel_gds_file->configure(-state=>"normal");
                                           $label_limitBox->configure(-state=>"disabled", -foreground => "$disable_fg");
                                           $entry_limitBox->configure(-state=>"disabled", -foreground => "$disable_fg");
                                         } else {
                                           $label_gds->configure(-state=>"disabled", -foreground => "$disable_fg");
                                           $gds_file->configure(-state=>"disabled", -foreground => "$disable_fg");
                                           $but_sel_gds_file->configure(-state=>"disabled");
                                           $label_limitBox->configure(-state=>"normal", -foreground => "$fg");
                                           $entry_limitBox->configure(-state=>"normal", -foreground => "$fg");
                                         }
                                        }
                       );




########################################################
# 
# ListBox to browse the rules containing errors.
# 
########################################################
my $listBox_frame = $main->Frame(
    )->pack(-side => 'top', -anchor => 'w', -fill => 'both', -expand => 1);

my $lb_rule_list = $listBox_frame->Scrolled("Listbox", -scrollbars => "e", -width => 30,
    -takefocus => 1, -selectmode => "extended"
    )->pack(-side => 'left', -fill => 'y', -expand => 0);

tie @rule_list_search, "Tk::Listbox", $lb_rule_list;

$lb_rule_list->bind('<ButtonPress>' => sub{ &fill_errors($lb_rule_list->curselection); });
$lb_rule_list->bind('<Key-Up>'      => sub{ &fill_errors($lb_rule_list->curselection); });
$lb_rule_list->bind('<Key-Down>'    => sub{ &fill_errors($lb_rule_list->curselection); });

my $text_desc;
sub fill_errors {
  my (@rule_sel_search) = @_ ;
  
  @rule_sel = ();
  if ( $search_pattern ne "" ) {
    foreach my $rule_sel_index (@rule_sel_search) {
      push(@rule_sel, $rule_list_index{@rule_list_search[$rule_sel_index]} );
    }
  } else { @rule_sel = @rule_sel_search }
  
  for ( my $k = 0; $k <= $#rule_list_search; $k++) {
    $lb_rule_list->itemconfigure( $k , -background => $bg );
  }
  if ( defined $lb_rule_list->curselection ) {
    $lb_rule_list->itemconfigure(@{$lb_rule_list->curselection}[0], -background => "grey95" );
  }
  
  @err_list = ();

  $cur_rule_layer = @rule_layer_list[@rule_sel[0]];

  my $j;
  my $k;
  my $error_array;
  my $coord_string;
  my $disp_ratio = 0;
  my $ratio;
  my $ratio_txt;
  my $min_ratio=100;
  my $max_ratio=0;
  if ( $dbase{$ENVI{TOP_CELL}}{ratio}{$cur_rule_layer}[0] ne "" ) { $disp_ratio = 1 }
  for ($j = 0; $j <= $#{$dbase{$ENVI{TOP_CELL}}{geom}{$cur_rule_layer}}; $j++) {
    $error_array = $dbase{$ENVI{TOP_CELL}}{geom}{$cur_rule_layer}[$j];
    $coord_string = "$j: ";
    $ratio=$dbase{$ENVI{TOP_CELL}}{ratio}{$cur_rule_layer}[$j] * 100;
    if ( $ratio < $min_ratio) { $min_ratio=$ratio }
    if ( $ratio > $max_ratio) { $max_ratio=$ratio }
    if ( $disp_ratio == 1 ) { $coord_string .= "(" . sprintf("%3.1f", $ratio) . ") ";}
    for ($k = 0; $k <= $#{$error_array}; $k++) { $coord_string .= $error_array->[$k] . " " }
    if      ( !($BUT_ratioFilter) || (($ENVI{RATIO_FILTER} >= 50) && ($ratio >= $ENVI{RATIO_FILTER})) ) {
      push(@err_list, $coord_string);
    } elsif ( !($BUT_ratioFilter) || (($ENVI{RATIO_FILTER}  < 50) && ($ratio <= $ENVI{RATIO_FILTER})) ) {
      push(@err_list, $coord_string);
    }
  }
  
  if ( $disp_ratio == 1 ) { $ratio_txt = "(min $min_ratio%, max $max_ratio%)" } else {$ratio_txt = ""}

  $err_name = $dbase{$ENVI{TOP_CELL}}{rule_nb}{$cur_rule_layer};
  $err_desc = $dbase{$ENVI{TOP_CELL}}{rule_name}{$cur_rule_layer};
  $text_desc->delete('1.0','end');
  $text_desc->insert("end", "$err_name : $err_desc $ratio_txt");

  
}



########################################################
# 
# ListBox to browse through the error list.
# Each error has a number. In case of ratio error type
# the ratio is add next to the error number.
# 
########################################################
my $lb_err_list = $listBox_frame->Scrolled("Listbox", -scrollbars => "e", -width => 30,
    -takefocus => 1, -selectmode => "browse"
    )->pack(-side => 'left', -fill => 'both', -expand => 1);


tie @err_list, "Tk::Listbox", $lb_err_list;

$lb_err_list->bind('<Button-1>' => sub{ &update_error } );
$lb_err_list->bind('<Key-Up>'   => sub{ &update_error } );
$lb_err_list->bind('<Key-Down>' => sub{ &update_error } );

sub update_error {

  &wait_dfII;

  if (! $dfII ) {&get_connection("localhost", $ENVI{SOCKET_PORT});}
  else          {open (SKILLTMP, ">$skilltmp") || die ("*** Error: Can't open $skilltmp !\n\n");}

  if ($BUT_ratioFilter) {
    my $error_string=@{$lb_err_list->get(@{$lb_err_list->curselection}[0],@{$lb_err_list->curselection}[0])}[0];
    $error_string =~ s/(\d+):.*/$1/;
    &disp_errors($error_string);
  } else { &disp_errors($lb_err_list->curselection); }

  if (! $dfII ) {close(SOCK) || die "Can't close Socket !";}
  else          {close(SKILLTMP) || die "Can't close $skilltmp !";}

  if ( $dfII ) {print "load(\"$skilltmp\") system(\"rm $skilltmp\") ";}
}

sub disp_errors {
  my ($err_sel,$hi_rule) = @_ ;
  
  my $j;
  my $start_point_x;
  my $start_point_y;
  my $end_point_x = $dbase{$ENVI{TOP_CELL}}{geom}{$cur_rule_layer}[$err_sel][0];
  my $end_point_y = $dbase{$ENVI{TOP_CELL}}{geom}{$cur_rule_layer}[$err_sel][1];
  my $min_x = $end_point_x;
  my $min_y = $end_point_y;
  my $max_x = $end_point_x;
  my $max_y = $end_point_y;
  my $mean_x;
  my $mean_y;
  
  if ( (! $dfII ) && $BUT_label ) {
    my $err_name = $dbase{$ENVI{TOP_CELL}}{rule_nb}{$cur_rule_layer};
    #dbCreateText replaced by add_text but the text size is now linked to the zooming factor. Asked CDN R&D to fix it.
    #print SOCK "dbCreateText -message $err_name -location $end_point_x $end_point_y\n";
    print SOCK "add_text -label $err_name -pt $end_point_x $end_point_y\n";
  }
  
  if ( ( $dfII ) && $BUT_label ) {
    my $err_name = $dbase{$ENVI{TOP_CELL}}{rule_nb}{$cur_rule_layer};
    print SKILLTMP "geAddHilightLabel(hlSetId $end_point_x:$end_point_y \"$err_name\" \"lowerLeft\" \"R0\" \"stick\" 2 nil) ";
  }

  for ($j = 3; $j <= $#{$dbase{$ENVI{TOP_CELL}}{geom}{$cur_rule_layer}[$err_sel]}; $j=$j+2) {
    $start_point_x=$end_point_x;
    $start_point_y=$end_point_y;
    $end_point_x = $dbase{$ENVI{TOP_CELL}}{geom}{$cur_rule_layer}[$err_sel][$j-1];
    $end_point_y = $dbase{$ENVI{TOP_CELL}}{geom}{$cur_rule_layer}[$err_sel][$j];
    if    ( $end_point_x   < $min_x ) {$min_x = $end_point_x}
    elsif ( $end_point_x   > $max_x ) {$max_x = $end_point_x}
    if    ( $end_point_y   < $min_y ) {$min_y = $end_point_y}
    elsif ( $end_point_y   > $max_y ) {$max_y = $end_point_y}
    
    if (! $dfII ) {
      print SOCK "
set dibvue_line_ptr [addCustomLine 1000 $start_point_x $start_point_y $end_point_x $end_point_y ]
lappend dibvue_line_list \$dibvue_line_ptr
";
    } else {
      print SKILLTMP "geAddHilightLine(hlSetId list($start_point_x:$start_point_y $end_point_x:$end_point_y)) ";
    }

  }

  if ( $hi_rule != 1 ) {
    if ( $BUT_autoZoom == 1 ) {
      if (! $dfII ) {
        print SOCK "zoomBox $min_x $min_y $max_x $max_y;zoomOut\n";
      } else {
        print SKILLTMP "hiZoomIn(cw${uid}_${pid} list($min_x:$min_y $max_x:$max_y)) hiZoomRelativeScale(cw${uid}_${pid} 0.5) ";
      }
    } elsif ( $BUT_pan == 1 ) {
      $mean_x = ($max_x + $min_x ) / 2;
      $mean_y = ($max_y + $min_y ) / 2;
      if (! $dfII ) {
        print SOCK "panCenter $mean_x $mean_y\n";
      } else {
        print SKILLTMP "hiPan(cw${uid}_${pid} list($mean_x $mean_y)) ";
      }
    } else {
      if (! $dfII ) {print SOCK "redraw\n";} else {print SKILLTMP "geRefreshCellView(cv${uid}_${pid}) ";} #print "hiRedraw(cw${uid}_${pid}) ";}
    }
  }
  
}

#   $entry_desc = $main->Entry(-text => "Rule description", -borderwidth => '1', -relief => 'ridge',
#    -font => "$font_name", -justify => 'left', -validate => 'all', -state => "readonly"
#    )->pack(-side => 'top',-expand => 'yes', -fill => 'x');

########################################################
# 
# Text widget that shows the rule comment of the selected rule.
# 
########################################################
   $text_desc = $main->Text(-borderwidth => '1', -relief => 'ridge',
    -font => "$font_name", -height => 4, -width => 60, -wrap => 'word'
    )->pack(-side => 'top',-expand => 'no', -fill => 'x');



########################################################
# 
# MainLoop
# 
########################################################

$checkBut_txt_file->invoke;
$checkBut_txt_file->invoke;
if ( $hercules == 1 ) { $checkBut_txt_file->invoke }
if ( $load == 1 )    { $but_load->invoke }

MainLoop;


################################################################################################################
################################################################################################################
#Subroutine section.
################################################################################################################


########################################################
# 
# subroutine scriptDialog($w, \$ent, \$path) => nil
# 
########################################################
sub scriptDialog {
    my $w = shift;
    my $ent = shift;
    my $path = shift;
    my $file;
    #   Type names        Extension(s)    Mac File Type(s)
    #
    #---------------------------------------------------------
    use File::Basename;
    my($filename, $directories) = fileparse($path);
    my @types =
      (["All files",            '*'],
       ["Calibre txt DB",       '.db'],
       ["Txt err files",        '.LAYOUT_ERRORS'],
       ["GDS files",            '.gds'],
      );
    $file = $w->getOpenFile(-filetypes => \@types, -initialdir => $directories );
    if (defined $file and $file ne '') {
      $ent->delete(0, 'end');
      $ent->insert(0, $file);
      $ent->xview('end');
    }
    return $file;
}

########################################################
# 
# subroutine but_editFile($widget, \$var) => nil
# 
########################################################
sub but_editFile {
    my ($widget, $var) = @_ ;
    my $file = $$var;
    my $design_dir = $ENVI{"DESIGN_DIR"};
    if ( -f "$design_dir/$file" ) {
        system "$ENVI{EDITOR} $design_dir/$file&";
    } elsif ( -f "$file" ) {
        system "$ENVI{EDITOR} $file&";
    } else {
my $dialog_noFile = $widget->Dialog( -title => "Message",
    -text => "File : $design_dir/$file not found!", -buttons => ["Ok"],
    -default_button => 'Ok',
);
$dialog_noFile->Show;
    }
} # sub but_editFile



###############################################################################
#
# Parsing the text error file
#
###############################################################################
sub dibvue_txt {
  
  print STDERR "Reading the txt File: $ENVI{TXT_FILE} ... ";
  
  open(FILE_TXT, $ENVI{TXT_FILE}) || die ("Error : Can't open txt file $ENVI{TXT_FILE} !\n\n");
  
  my $rule_name = "";
  my $rule_nb;
  my $layer;
  my $layer_data;
  my $datatype;
  my $parse_error_list;
  my $err_count;
  my $parse_ratio;
  my $tmp;
  my $fill_err_txt;
  
  my $top = $ENVI{TOP_CELL};
  print STDERR "\n-I- top cell (from GDS) : -->$ENVI{TOP_CELL}<--\n";
  
  while(<FILE_TXT>) {
    ########################################################
    # 
    # Parsing the txt file ...
    # 
    ########################################################
    
    if ( $top eq "" ) {
      ########################################################
      # 
      # If the top cell name was not found inside the GDS
      # then tries to find it inside the TXT errro file.
      ########################################################
      if ( /Structure name: +(\S+)/ ) {
        $top = $1;
        $ENVI{TOP_CELL} = $1;
        print STDERR "\n-I- top cell (from TXT) : -->$ENVI{TOP_CELL}<--\n";
      }
    }

    if ( /^$top +(\S+) +(\S+) +(\(.*\)) +(\(.*\)) +(\d+);(\d+)/ ) {
      ########################################################
      # 
      # Handles the "ERR_TEXT_SHORTEST_PATH" section if
      # FIND_SHORTEST_PATH_BETWEEN_TEXT_SHORTS = 100 is set.
      ########################################################
      $rule_nb = "SHORT_$1_$2";
      $rule_name = "Short between $1 $3 and $2 $4";
      $layer=$5;
      $datatype=$6;
      $dbase{$top}{rule_nb}{"$layer:$datatype"}=$rule_nb;
      $dbase{$top}{rule_name}{"$layer:$datatype"}=$rule_name;
    }
    
    #-- Finding the text violation count in the text header section.
    if ( /\((\d+);(\d+)\) +\.* +(\d+) violation(s)* found/ ) { $dbase{$top}{err_count_txt}{"$1:$2"} = $3;}
    
    if ( /#####--- ERR/ ) { $rule_name = ""; $parse_error_list = 0;}
  
    #-- Identifying the rules numbers and description for standard rules.
    if ( /\tCOMMENT = \"( *Rule *)*(\S+) *(: *)* *(.*)\"/ ) { $rule_nb = $2; $rule_name = $4; }
  
    #-- Identifying the standard rules GDS layers. Checking if GDS data is present.
    if ( /} *\((\d+);(\d+)\)/ && ($rule_name ne "") ) {
      $layer=$1;
      $datatype=$2;
      $dbase{$top}{rule_nb}{"$layer:$datatype"}=$rule_nb;
      $dbase{$top}{rule_name}{"$layer:$datatype"}=$rule_name;
      if (! defined @{$dbase{$top}{geom}{"$layer:$datatype"}} ) {
        print STDERR "-W- No GDS markers for $rule_nb, will use TXT markers if any instead !\n";
        $fill_err_txt = 1;
        $dbase{$top}{rule_nb}{"$layer:$datatype"}="$rule_nb (noGDS!)"
      }
    }
    
    if ( /^Structure +/ ) {
      $parse_error_list = 1;
      $err_count = 0;
      if ( /^Structure +.*Ratio/ ) { 
        $parse_ratio = 1;
        #-- In case of errors limit to 1000 the density matching (ratio filtering) is not possible.
        if ( $dbase{$top}{err_count_txt}{"$layer:$datatype"} == 1000 ) {
          print STDERR "-W- Warning. Error limit 1000 for the error $rule_nb. The Ratio filter won't be available for this rule !!!\n";
          $parse_ratio = 0;
        }
      }
    }
    
    if ( ($rule_name ne "") && ($parse_error_list == 1) && (/^$top/) ) {
      if ( ($parse_ratio == 1) || (/^$top .*Density/) ) {
        ########################################################
        # 
        # The ratio found in the txt file needs to be matched to
        # the right error in the GDS. Matching done by comparing
        # the first coordinate of the square in the wo files.
        # 
        ########################################################
        if ( /^$top +\((-*\d+\.\d+), +(-*\d+\.\d+).*\) +(Density[1]* = )*(\S+) +/ ) {
          my $txt_x = sprintf("%20.5f", $1);
          my $txt_y = sprintf("%20.5f", $2);
          my $gds_x;
          my $gds_y;

          my $j;
          my $dens_match=0;
          for ($j = 0; $j <= $#{$dbase{$ENVI{TOP_CELL}}{geom}{"$layer:$datatype"}}; $j++) {
            $gds_x = sprintf("%20.5f", $dbase{$ENVI{TOP_CELL}}{geom}{"$layer:$datatype"}[$j][0]);
            $gds_y = sprintf("%20.5f", $dbase{$ENVI{TOP_CELL}}{geom}{"$layer:$datatype"}[$j][1]);
            if ( ($gds_x == $txt_x) && ($gds_y == $txt_y) ) {
              if ($dens_match) { print STDERR "-W- Density matching already found !!!\n"}
              $dbase{$top}{ratio}{"$layer:$datatype"}[$j] = $4;
              $dens_match=1;
            }
          }
          if ($dens_match == 0) { print STDERR "-E- Density matching not found ($txt_x, $txt_y), $layer:$datatype, $4\n"}
        }
        else { die ("Error : Can't find ratio for rule $rule_nb error nb $err_count !\n\n")}
      }

      #-- Using the text markers in case of missing geometry data in the GDS.
      if ( ($fill_err_txt == 1) && (/^$top +\((-*\d+\.\d+), +(-*\d+\.\d+).*[\) \(|, ](-*\d+\.\d+), +(-*\d+\.\d+).*\)/) ) {
        push(@{$dbase{$top}{geom}{"$layer:$datatype"}},[$1, $2, $1, $4, $3, $4, $3, $2, $1, $2]);
      }
      
      $err_count++;
    }
    
    if ( ($rule_name ne "") && ($parse_error_list == 1) && (/^$/) ) {
      $parse_error_list =0;
      $parse_ratio = 0;
      $fill_err_txt = 0;
      #$dbase{$top}{err_count_txt}{"$layer:$datatype"} = $err_count;
      my $err_count_string = "(" . $err_count . ")";
      my $err_flag;
      if ($dbase{$top}{err_count_txt}{"$layer:$datatype"} == $err_count) {$err_flag = " "} else {$err_flag = "*"}
      printf STDERR "Nb err.: %6d%-8s %s Rule : %15s  --> %s\n", $dbase{$top}{err_count_txt}{"$layer:$datatype"}, $err_count_string, $err_flag, $rule_nb, $rule_name;

    }
  
  }
  #-- Bug fix to print the summary for the last error when it has not been display in the While loop above.
  if ( ($rule_name ne "") && ($parse_error_list == 1) ) {
    my $err_count_string = "(" . $err_count . ")";
    my $err_flag;
    if ($dbase{$top}{err_count_txt}{"$layer:$datatype"} == $err_count) {$err_flag = " "} else {$err_flag = "*"}
    printf STDERR "Nb err.: %6d%-8s %s Rule : %15s  --> %s\n", $dbase{$top}{err_count_txt}{"$layer:$datatype"}, $err_count_string, $err_flag, $rule_nb, $rule_name;

  }

  print STDERR "done\n";

  close(FILE_TXT) || die "Can't close $ENVI{TXT_FILE} !";
  
}


###############################################################################
#
# Parsing the GDS error file
#
###############################################################################
sub dibvue_gds {
  
  print STDERR "\nReading the gds File: $ENVI{GDS_FILE} ... \n";

  $ENVI{TOP_CELL}="--------";

  use GDS2;
  my $gds2File = new GDS2(-fileName=>$ENVI{GDS_FILE});
  my $cell_name;
  my $layer;
  my $datatype;
  my @xy;
  my $sref_cell;
#  my $gds_error_limit;
  my $angle;
  
  while ($gds2File -> readGds2Record) {
    
    ########################################################
    # 
    # Finding the top cell_name
    # 
    ########################################################
    if ( $gds2File -> isStrname )  {
      if ($gds2File -> returnStrname =~ /EV_OUTPUT/) {
        $gds2File -> readGds2Record;
        $gds2File -> readGds2Record;
        if ( $gds2File -> isSname ) { $ENVI{TOP_CELL} = $gds2File -> returnSname; $ENVI{TOP_CELL} =~ s/^ERR_//; }
      }
    }

    ########################################################
    # 
    # Finding the current cell_name
    # 
    ########################################################
    if ( $gds2File -> isStrname )  {
      if ($gds2File -> returnStrname =~ /^ERR_/) {
        $cell_name = $gds2File -> returnStrname;
        $cell_name =~ s/^ERR_//;
      }
    }

    ########################################################
    # 
    # Parsing an error Polygon with it's layer:datatype and coordinates
    # 
    ########################################################
    if ( $gds2File -> isBoundary ) {
      $gds2File -> readGds2Record;
      if ( $gds2File -> isLayer ) {
        $layer     = $gds2File -> returnLayer;
        $gds2File -> readGds2Record;
        if ( $gds2File -> isDatatype ) {
          $datatype  = $gds2File -> returnDatatype;
          $gds2File -> readGds2Record;
          if ( $gds2File -> isXy ) {
            @xy = $gds2File -> returnXyAsArray(-asInteger => 0);
            $gds2File -> readGds2Record;
            if ($gds2File -> isEndel) { 
#              my $count_gds = $dbase{$cell_name}{err_count_gds}{"$layer:$datatype"}++;
#              if ( $count_gds < 1300 ) {
#                push(@{$dbase{$cell_name}{geom}{"$layer:$datatype"}},[@xy]);
#              } else { $gds_error_limit = 1 }
              push(@{$dbase{$cell_name}{geom}{"$layer:$datatype"}},[@xy]);
            } else { print STDERR "-E- GDS Error : Missing ENDEL in BOUNDARY record !\n\n";}
          } else   { print STDERR "-E- GDS Error : Missing XY array in BOUNDARY record !\n\n";}
        } else     { print STDERR "-E- GDS Error : Missing datatype in BOUNDARY record !\n\n";}
      } else       { print STDERR "-E- GDS Error : Missing layer in BOUNDARY record !\n\n";}
    }
    
    ########################################################
    # 
    # Parsing SREFs for sub hier cells positions
    # 
    ########################################################
    if ( $gds2File -> isSref ) {
      if ( $cell_name eq $ENVI{TOP_CELL} ) {
        $gds2File -> readGds2Record;
        if ( $gds2File -> isSname ) {
          $sref_cell = $gds2File -> returnSname;
          $sref_cell =~ s/^ERR_//;
          $gds2File -> readGds2Record;
          if ( $gds2File -> isStrans ) {
            $gds2File -> readGds2Record;
            $angle = 0;
            if ( $gds2File -> isAngle ) {
              $angle = $gds2File -> returnRecordAsString;
              $angle =~ s/^\s+ANGLE\s+(\d+)/\1/;
              $gds2File -> readGds2Record;
            }
            if ( $gds2File -> isXy ) {
              @xy = $gds2File -> returnXyAsArray(-asInteger => 0);
              $gds2File -> readGds2Record;
              if ($gds2File -> isEndel) {
                if ( $angle == 0 ) {
                  #print STDERR "SREF --$sref_cell-- --@xy--\n";
                  push(@{$dbase{$sref_cell}{instances}},[@xy]);
                } else { print STDERR "-W- GDS Warning : ANGLE different from 0 ($angle) not supported in SREF $sref_cell !\n\n"; }
              } else { print STDERR "-E- GDS Error : Missing ENDEL in SREF record !\n\n";}
            } else   { print STDERR "-E- GDS Error : Missing XY array in SREF record !\n\n";}
          } else     { print STDERR "-E- GDS Error : Missing STRANS in SREF record !\n\n";}
        } else       { print STDERR "-E- GDS Error : Missing SNAME in SREF record !\n\n";}
      }else          { print STDERR "-W- GDS Warning : dibvue is supporting only one level of hierarchy in SREF records !\n\n";}
    }

    ########################################################
    # 
    # Parsing AREFs for sub hier cells positions
    # 
    ########################################################
    my $colrow;
    my @colrow;
    my $offset_x;
    my $offset_y;
    my $pitch_x;
    my $pitch_y;
    my $i;
    my $j;
    if ( $gds2File -> isAref ) {
      $gds2File -> readGds2Record;
      if ( $gds2File -> isSname ) {
        $sref_cell = $gds2File -> returnSname;
        $sref_cell =~ s/^ERR_//;
        $gds2File -> readGds2Record;
        if ( $gds2File -> isStrans ) {
          $gds2File -> readGds2Record;
          if ( $gds2File -> isColrow ) {
            $colrow = $gds2File -> returnRecordAsString;
            $colrow =~ s/^\s+//;
            @colrow = split(/\s+/, $colrow );
            $gds2File -> readGds2Record;
            if ( $gds2File -> isXy ) {
              @xy = $gds2File -> returnXyAsArray(-asInteger => 0);
              $gds2File -> readGds2Record;
              if ($gds2File -> isEndel) {
                #print STDERR "AREF --$sref_cell-- --@colrow-- --@xy--\n";
                $offset_x = @xy[0];
                $offset_y = @xy[1];
                $pitch_x= (@xy[2] - @xy[0]) / @colrow[1];
                $pitch_y= (@xy[5] - @xy[1]) / @colrow[2];
                for ($i = 0; $i < @colrow[1]; $i++) {
                  for ($j = 0; $j < @colrow[2]; $j++) {
                    @xy = ( $offset_x + $i * $pitch_x , $offset_y + $j * $pitch_y );
                    push(@{$dbase{$sref_cell}{instances}},[@xy]);
                  }
                }
              } else { print STDERR "-E- GDS Error : Missing ENDEL in BOUNDARY record !\n\n";}
            } else   { print STDERR "-E- GDS Error : Missing XY array in BOUNDARY record !\n\n";}
          } else     { print STDERR "-E- GDS Error : Missing COLROW in BOUNDARY record !\n\n";}
        } else       { print STDERR "-E- GDS Error : Missing STRANS in BOUNDARY record !\n\n";}
      } else         { print STDERR "-E- GDS Error : Missing SNAME in BOUNDARY record !\n\n";}
    }



  }

#  if ( $gds_error_limit == 1 ) { print STDERR "-W- Warning, maximum number of error per rule reached for one rule at least : 50000\n" }

  ########################################################
  # 
  # Translating sub cells errors in the top cell coordinates.
  # 
  ########################################################
  my @cell_list;
  my @xy_tmp;
  my $cell;
  my @sub_rule_nb_list;
  my $rule_layer;
  my $instance;
  my $i;
  my $j;
  @cell_list = keys(%dbase);
  foreach $cell (@cell_list) {
    if ( $cell ne $ENVI{TOP_CELL} ) {
      @sub_rule_nb_list = keys(%{$dbase{$cell}{geom}});
      foreach $instance (@{$dbase{$cell}{instances}}) {
        foreach $rule_layer (@sub_rule_nb_list) {
          for ($i = 0; $i <= $#{$dbase{$cell}{geom}{$rule_layer}}; $i++) {
            @xy = @{$dbase{$cell}{geom}{$rule_layer}[$i]};
            for ($j = 1; $j <= $#xy; $j=$j+2) {
              @xy_tmp[$j-1] = @xy[$j-1] + $instance->[0];
              @xy_tmp[$j]   = @xy[$j]   + $instance->[1];
            }
            push(@{$dbase{$ENVI{TOP_CELL}}{geom}{$rule_layer}},[@xy_tmp]);
          }
        }
      }
    }
  }

  print STDERR "done\n";
  
}

###############################################################################
#
# Parsing the text error file for Calibre
#
###############################################################################

my $prev_line;
my $line;

sub dibvue_calibre {
  
  print STDERR "Reading the Calibre db text File: $ENVI{TXT_FILE} ... ";
  
  open(FILE_TXT, $ENVI{TXT_FILE}) || die ("Error : Can't open txt file $ENVI{TXT_FILE} !\n\n");

  my $rule="???";
  my $text;
  my $nb;
  my $text_line_nb;
  my $j;
  my @xy;
  my $rule_nb_warn;
  my $parse_geom=0;
  my $geom_type;
  my $geom_nb;
  my $geom_count;
  my $gcount;
  my $line_nb=1;

  my $top = <FILE_TXT>;
  chomp($top);
  $top =~ s/^(\S+).*$/\1/;
  $ENVI{TOP_CELL}=$top;
  print STDERR "\n-I- top cell : -->$ENVI{TOP_CELL}<--\n";

  while( $line = <FILE_TXT> ) {
    ########################################################
    # 
    # Parsing the txt file ...
    # 
    ########################################################
    $line_nb++;
    chomp($line);

    #Looking for the DATE line with the error count at the beginning of each rule.
    if ( $line =~ /(\d+) +\d+ +(\d+) +\S+ +\d+ +\d+:\d+:\d+ +\d+/ ) {
      if ( $parse_geom == 1 ) {
        printf STDERR "-E- Error (line %d): missing coordinates for $geom_type $geom_nb rule $rule: expected $geom_count, found %d ! bad text file?\n", $line_nb-1, $geom_count-$gcount;
        $parse_geom=0;
      }
      $rule_nb_warn = 0;
      $rule = $prev_line;
      $text="";
      ($nb,$text_line_nb) = ($line =~ /^(\d+) \d+ (\d+) /);
      if ( $text_line_nb == 0 ) { $text  = $rule }
      for ($j= 1; $j <= $text_line_nb; $j++) {
        $line = <FILE_TXT>;
        $line_nb++;
        chomp($line);
        $line =~ s/ +$//;
        if ( $line !~ /Rule File Pathname/ ) {$text .= $line . " ";}
      }
      printf STDERR "Nb err.: %6d  Rule : %15s  --> %s\n", $nb, $rule, $text;
      if ($nb >0) {
        $dbase{$top}{rule_nb}{"ID_$rule"}=$rule;
        $dbase{$top}{rule_name}{"ID_$rule"}=$text;
      }
    }

    #Looking for a polygon or edge header.
    if ( $line =~ /^(p|e) (\d+) (\d+)/ ) {
      if ( $parse_geom == 1 ) {
        printf STDERR "-E- Error (line %d): missing coordinates for $geom_type $geom_nb rule $rule: expected $geom_count, found %d ! bad text file?\n", $line_nb-1, $geom_count-$gcount;
      }
      $geom_type=$1;
      $geom_nb=$2;
      $geom_count=$3;
      $gcount=$3;
      if ( ($geom_nb > $nb) && ($rule_nb_warn == 0) ) {$rule_nb_warn = 1; printf STDERR "-W- More shapes than specified in the rule header ($nb) for rule $rule !\n"}
      @xy=();
      if ( $geom_nb <= $ENVI{LIMIT} ) {
        $parse_geom=1;
      }
    }
      
    #Parsing the polygon or edge coordinates.
    if ( $parse_geom == 1 ) {
      if ( $line =~ /^-*\d+( +-*\d+)+$/ ) {
        #Find lines that are polygon coordinates lists
        chomp($line);
        push(@xy, split(/\s+/,$line));
        $gcount--;
      }
      if ( $gcount == 0 ) {
        #Parsed all the coordinates of a polygon or edge, adding it to the DB.
        $parse_geom = 0;
        if ($geom_type eq "p") {push(@xy, $xy[0], $xy[1]);}
        for ($j = 0; $j <= $#xy; $j++) { $xy[$j]=($xy[$j]*$magnify)/1000;}
        push(@{$dbase{$top}{geom}{"ID_$rule"}},[@xy]);
      }
    } else {
      #Issue an Error message if coordinates are lost and discarded from the DB.
      if ( ( $geom_nb <= $ENVI{LIMIT} ) && ($line =~ /^-*\d+( +-*\d+)+$/) ) {
        printf STDERR "-E- Error (line $line_nb): more coordinates found than expected $geom_count for rule $rule, this coordinate is lost!\n";
      }
    }
        
    $prev_line = $line;
  }
  print STDERR "done\n";

  close(FILE_TXT) || die "Can't close $ENVI{TXT_FILE} !";
}


